from typing import Dict, TypeVar
from random import choices


T = TypeVar('T')

class PMF:
    _pmf: Dict[T, int]
    _total_weight: int

    def __init__(self, items_with_weights: Dict[T, int] = None):
        self._pmf = dict()
        self._total_weight = 0
        if items_with_weights != None:
            self.add_items(items_with_weights)

    def items(self):
        return self._pmf.keys()

    def add_items(self, items_with_weights: Dict[T, int], update_existing=False):
        for key in items_with_weights:
            temp_weight = items_with_weights[key]
            if (not update_existing) and (key in self._pmf):
                raise KeyError("item {} already exists in PMF!".format(key))
            else:
                self._total_weight -= self._pmf.get(key, 0.0)
                self._pmf[key] = temp_weight
                self._total_weight += temp_weight

    def delete_item(self, item: T):
        if item in self._pmf:
            temp_weight = self._pmf[item]
            del self._pmf[item]
            self._total_weight -= temp_weight

    def draw_n(self, n: int, replace=True):
        items = []
        probs = []
        total_weight = float(self._total_weight)
        for key in self._pmf:
            items.append(key)
            probs.append((self._pmf[key] / total_weight))
        return choices(items, weights=probs, k=n)

    def get_total_weight(self):
        return self._total_weight


from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse

class BestRespondAgent(LemonadeAgent):
    def setup(self):
        self.next_action = 0
    
    def get_action(self):
        return self.next_action

    def update(self):
        l1 = self.get_opp1_last_action()
        l2 = self.get_opp2_last_action()
        if l1 == l2:
            self.next_action = (l1 + 6) % 12
        else:
            tmp = None
            if l1 < l2:
                tmp = l1
                l1 = l2
                l2 = tmp
            if (l1 - l2) > (12 - (l1 - l2)):
                self.next_action = (l1 - ((l1 + l2) // 2)) % 12
            else:
                self.next_action = (l1 + ((12 - (l1 - l2)) // 2)) % 12

################### SUBMISSION #####################
agent_submission = BestRespondAgent("Best Response")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    # parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent_submission,
                BestRespondAgent("Agent_1"),
                BestRespondAgent("Agent_2"),
                BestRespondAgent("Agent_3"),
                BestRespondAgent("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.lsvm_agent import MyLSVMAgent
from agt_server.local_games.lsvm_arena import LSVMArena
from agt_server.agents.test_agents.lsvm.min_bidder.my_agent import MinBidAgent
from agt_server.agents.test_agents.lsvm.jump_bidder.jump_bidder import JumpBidder
from agt_server.agents.test_agents.lsvm.truthful_bidder.my_agent import TruthfulBidder
import time
import random
from .path_utils import path_from_local_root


NAME = "TA - RANDOM"
class MyAgent(MyLSVMAgent):
    def setup(self):
        self.filename = path_from_local_root("test.txt")
        self.file = open(self.filename, 'r+')

    def get_bids(self):
        min_bids = self.get_min_bids()
        valuations = self.get_valuations() 
        bids = {}
        weight = int(self.file.read(1))
        self.file.seek(0)
        for good in valuations: 
            if valuations[good] > min_bids[good]:
                bids[good] = min_bids[good] + (weight * (valuations[good] - min_bids[good]))
        return bids
    
    def update(self):
        weight = random.random()
        self.file.write(str(weight))
        self.file.truncate()
        self.file.seek(0)

################### SUBMISSION #####################
my_agent_submission = MyAgent(NAME)
####################################################

if __name__ == "__main__":
    ### DO NOT TOUCH THIS #####
    agent = MyAgent(NAME)
    arena = LSVMArena(
        num_cycles_per_player = 3,
        timeout=1,
        # local_save_path="saved_games",
        players=[
            agent,
            MyAgent("CP - MyAgent"),
            MyAgent("CP2 - MyAgent"),
            MyAgent("CP3 - MyAgent"),
            MinBidAgent("Min Bidder"), 
            JumpBidder("Jump Bidder"), 
            TruthfulBidder("Truthful Bidder"), 
        ]
    )
    
    start = time.time()
    arena.run()
    end = time.time()



from typing import Set, List, Dict
import itertools

class MarketSegment(frozenset):
    _all_segments = [] 

    def __new__(cls, iterable):
        return super(MarketSegment, cls).__new__(cls, iterable)
    
    def __init__(self, iterable):
        super().__init__()
        self.name = "_".join(iterable)
        
    @classmethod
    def all_segments(cls):
        return cls._all_segments

MarketSegment._all_segments = [
    MarketSegment(("Male", "Young")),
    MarketSegment(("Male", "Old")),
    MarketSegment(("Male", "LowIncome")),
    MarketSegment(("Male", "HighIncome")),
    MarketSegment(("Female", "Young")),
    MarketSegment(("Female", "Old")),
    MarketSegment(("Female", "LowIncome")),
    MarketSegment(("Female", "HighIncome")),
    MarketSegment(("Young", "LowIncome")),
    MarketSegment(("Young", "HighIncome")),
    MarketSegment(("Old", "LowIncome")),
    MarketSegment(("Old", "HighIncome")),
    MarketSegment(("Male", "Young", "LowIncome")),
    MarketSegment(("Male", "Young", "HighIncome")),
    MarketSegment(("Male", "Old", "LowIncome")),
    MarketSegment(("Male", "Old", "HighIncome")),
    MarketSegment(("Female", "Young", "LowIncome")),
    MarketSegment(("Female", "Young", "HighIncome")),
    MarketSegment(("Female", "Old", "LowIncome")),
    MarketSegment(("Female", "Old", "HighIncome"))
]

class Bid:
    _uid_generator = itertools.count(1)
    
    uid: int
    bidder: any
    auction_item_spec: MarketSegment
    bid_per_item: float
    bid_limit: float # same units as bid_per_item (e.g. dollars)

    def __init__(self, bidder, auction_item: Set[str], 
                       bid_per_item: float = 0.0, bid_limit: float = 0.0):
        self.uid = next(type(self)._uid_generator)
        self.bidder = bidder
        self.item = auction_item
        self.bid_per_item = bid_per_item
        self.bid_limit = bid_limit
        
        # A bid cannot be negative
        try:
            assert self.bid_per_item >= 0
        except:
            print("bpi: {}".format(self.bid_per_item))
        # A limit cannot be non-positive 
        assert self.bid_limit >= 0
        # A bid cannot be bigger than its limit, since in the worst case a bidder could end up paying a price arbitrarily close to its bid.
        assert self.bid_per_item <= self.bid_limit

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.uid == other.uid

    def __hash__(self):
        return hash(self.__class__.__name__ + str(self.uid))

    def __repr__(self):
        return "{}(uid: {}, bidder: {}, auction_item_spec: {})".format(self.__class__.__name__, 
                                                                       self.uid, self.bidder, 
                                                                       self.auction_item_spec.uid)

    def deduct_limit(self, price: float):
        self.total_limit -= price

    @classmethod
    def from_vector(cls, bid_vec, bidder, auction_item_spec):
        '''
        Assuming bid vector is: [auction_item_spec_id, bid_per_item, total_limit].
        '''
        bid_per_item = bid_vec[1]
        total_limit = bid_vec[2]
        return cls(bidder, auction_item_spec, bid_per_item, total_limit)

    def to_vector(self):
        return [self.auction_item_spec.uid, self.bid_per_item, self.total_limit]

class Campaign:
    """
    Represents a campaign
    """
    _uid_generator = itertools.count(1)
    
    uid: int
    _reach: int
    _budget: float
    _target: MarketSegment # What segment to target

    def __init__(self, reach, target, start_day, end_day):
        self.uid = next(type(self)._uid_generator)
        self._reach = reach
        self._budget = None
        self._target = target
        self._start = start_day
        self._end = end_day
        self._costs = 0 
        self._impressions = 0
    
    @property 
    def start_day(self):
        return self._start
    
    @property 
    def end_day(self):
        return self._end

    @property
    def reach(self):
        return self._reach

    @property
    def budget(self):
        return self._budget

    @budget.setter 
    def budget(self, value):
        if self._budget is None:
            self._budget = value

    @property
    def target_segment(self):
        return self._target
    
    @property 
    def start(self):
        return self._start 
    
    @property 
    def end(self):
        return self._end
    
    @property 
    def cumulative_reach(self):
        return self._impressions 
    
    @cumulative_reach.setter 
    def cumulative_reach(self, impressions):
        self._impressions = impressions

    
    @property 
    def cumulative_cost(self):
        return self._costs

    @cumulative_cost.setter 
    def cumulative_cost(self, costs):
        self._costs = costs

    def __repr__(self):
        return "{}(uid: {}, reach: {}, budget: {}, target: {})".format(self.__class__.__name__,
                                                                       self.uid, self.reach, 
                                                                       self.budget, self.target)

    def __lt__(self, other):
        return (self.budget / self.reach) <= (other.budget / other.reach)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.uid == other.uid

    def __hash__(self):
        return hash(self.__class__.__name__ + str(self.uid))
    
class BidBundle:
    campaign_id: int 
    limit: float 
    bid_entries: Set[Bid]

    def __init__(self, campaign_id : int, limit : float, bid_entries : Set[Bid]):
        self.campaign_id = campaign_id 
        self.limit = limit 
        self.bid_entries = bid_entries 

    def __repr__(self):
        return f"BidBundle(id={self.campaign_id}, limit={self.limit}, bid_entries={self.bid_entries})"




from agt_server.agents.base_agents.lsvm_agent import MyLSVMAgent
from agt_server.local_games.lsvm_arena import LSVMArena
from agt_server.agents.test_agents.lsvm.min_bidder.my_agent import MinBidAgent
from agt_server.agents.test_agents.lsvm.jump_bidder.jump_bidder import JumpBidder
from agt_server.agents.test_agents.lsvm.truthful_bidder.my_agent import TruthfulBidder
import time


NAME = "TA - SECRET"
class MyAgent(MyLSVMAgent):
    def setup(self):
        self.lastNumbRounds = 650
        self.curRoundNumb = 0
        self.notBidding = set() 
    
    def get_space_valuation(self, aloc):
        val = 0
        for i in range(6):
            for j in range(3):
                temp = self.get_space_score(aloc, i, j)
                val += pow(temp, 1.5)
        return val
    
    def get_real_val(self, aloc, roundsRemaining):
        rx = roundsRemaining / self.lastNumbRounds
        spaceVal = self.get_space_valuation(aloc)
        val = self.calc_total_utility(aloc)

        delta = 1 - rx
        return val + delta * spaceVal

    def get_space_score(self, aloc, x, y):
        val = 0
        for a in aloc:
            loc = self.get_location(a, 6)
            if loc[0] == x:
                if loc[1] == y + 1 or loc[1] == y - 1:
                    val += 1
            elif loc[1] == y:
                if loc[0] == x + 1 or loc[0] == x - 1:
                    val += 1
        return val
    
    def get_location(self, s, numC):
        val = ord(s[0]) - ord('A')
        return [val % numC, val // numC]
    
    def bid_func(self, roundsRemaining, s):
        minBids = self.get_min_bids()
        fval = 0.8
        val = 0
        if self.get_current_round() == 0: 
            if self.get_valuation(s) * fval > minBids[s]: 
                val = self.get_valuation(s) * fval
            else:
                val = minBids[s]
        elif s not in self.notBidding:
            currAloc = self.get_tentative_allocation() 
            currAloc = set(currAloc)
            if s in currAloc:
                currAloc.remove(s)
            curVal = self.get_real_val(currAloc, roundsRemaining)
            currAloc.add(s)
            if self.get_real_val(currAloc, roundsRemaining) > curVal + minBids[s]:
                if minBids[s] + (self.get_real_val(currAloc, roundsRemaining) - curVal - minBids[s] * fval) > minBids[s]:
                    val = minBids[s] + (self.get_real_val(currAloc, roundsRemaining) - curVal - minBids[s] * fval)
                else:
                    val = minBids[s]
            else:
                self.notBidding.add(s)
        return val

    def national_bidder(self, roundsRemaining, minBids):
        ret = {}
        for s in minBids.keys():
            val = self.bid_func(roundsRemaining, s)
            if val > 0:
                ret[s] = val
        return ret
    
    def get_bids(self):
        minBids = self.get_min_bids()
        roundsRemaining = self.lastNumbRounds - self.curRoundNumb
        roundsRemaining = max(roundsRemaining, 1)
        ret = {}
        if self.is_national_bidder():  # Assuming this method is defined elsewhere
            ret = self.national_bidder(roundsRemaining, minBids)
        else:
            ret = self.national_bidder(roundsRemaining, minBids)
        return ret
    
    def update(self):
        self.curRoundNumb += 1

################### SUBMISSION #####################
my_agent_submission = MyAgent(NAME)
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    agent = MyAgent(NAME)
    arena = LSVMArena(
        num_cycles_per_player = 3,
        timeout=1,
        # local_save_path="saved_games",
        players=[
            agent,
            MyAgent("CP - MyAgent"),
            MyAgent("CP2 - MyAgent"),
            MyAgent("CP3 - MyAgent"),
            MinBidAgent("Min Bidder"), 
            JumpBidder("Jump Bidder"), 
            TruthfulBidder("Truthful Bidder"), 
        ]
    )
    
    start = time.time()
    arena.run()
    end = time.time()



import numpy as np
from agt_server.agents.base_agents.lsvm_agent import MyLSVMAgent
from agt_server.local_games.lsvm_arena import LSVMArena
from agt_server.agents.test_agents.lsvm.min_bidder.my_agent import MinBidAgent
from agt_server.agents.test_agents.lsvm.jump_bidder.jump_bidder import JumpBidder
from agt_server.agents.test_agents.lsvm.truthful_bidder.my_agent import TruthfulBidder
import time
import os
import gzip
import json
from .path_utils import path_from_local_root
import random

NAME = "PW - Hot'n'Cold"

class MyAgent(MyLSVMAgent):
    def setup(self):
        """
        - Bid aggressively on items where we have a high true value relative to the distribution; otherwise bid passively
        - Identify whether an item on the grid has an aggressive bidder or not by bidding a medium-range jump bid; this will 
            beat out all passive bidders but not any aggressive bidders, so if we are not winning the item in round 1 then 
            there must be an aggressive bidder
        - When we reach close to our true item values in the items with the aggressive bidders, we then decide whether we 
            attempt to outbid them; if removing this item from that bidder's winnings (based on the current tentative 
            allocations) would reduce the aggressive bidders' value significantly, we know this would split a region of the 
            opponent's winning bundle into 2 smaller regions that aren't connected, so we start to bid more aggressively to 
            try and sabotage that opponent
        """
        # Define variables relating to information-gathering
        self.is_opponent_aggressive = dict()
        self.play_aggressive = dict()
        
        # Define thresholds for us to choose aggressive versus passive bidding
        self.distribution_threshold_national = 7
        self.distribution_threshold_regional = 9
        self.passive_bidding_factor = 1.1
        self.aggressive_bidding_factor_national = 4
        self.aggressive_bidding_factor_regional = 2.5
        self.aggressive_initial_jump_bid_factor = 0.95

        # Define thresholds to identify aggressive versus passive opponents
        self.initial_bid_national = 4
        self.initial_bid_regional = 4

        # Define thresholds to swap to aggressive bidding for spatial reasons
        self.self_swap_factor = 1.5
        self.opponent_swap_factor = 2
        self.min_items_in_bundle = 4

    def national_bidder_strategy(self):
        min_bids = self.get_min_bids()
        valuations = self.get_valuations(self.get_goods())
        goods = self.get_goods()
        tentative_allocation_goods = self.get_tentative_allocation()
        prior_bids = self.get_previous_bid_map()

        bids = dict()

        # Iterate through each item
        for good in goods:
            if self.get_current_round() == 0:
                # Bid the initial bid to get information, or aggressively if we have a high value
                if valuations[good] >= self.distribution_threshold_national:
                    self.play_aggressive[good] = True
                    bids[good] = self.aggressive_initial_jump_bid_factor * valuations[good]
                else:
                    self.play_aggressive[good] = False
                    bids[good] = min(self.initial_bid_national, valuations[good])

            else:
                if self.get_current_round() == 1:
                    # Gather information about aggressive vs passive for each item right after first round
                    if good not in tentative_allocation_goods:
                        self.is_opponent_aggressive[good] = True
                    else:
                        self.is_opponent_aggressive[good] = False
                        self.play_aggressive[good] = True

                # If we already own the good, hold our bid
                if good in tentative_allocation_goods and good in prior_bids:
                    bids[good] = prior_bids[good]
                
                else:
                    # Verify if we want to swap to an aggressive strategy to block off an opponent
                    prior_winners = self.get_previous_winners_map()
                    curr_item_prior_winner = prior_winners[good]
                    prior_winner_bundle = set()

                    for winners_good in prior_winners:
                        if prior_winners[winners_good] == curr_item_prior_winner:
                            prior_winner_bundle.add(winners_good)
                    
                    curr_winner_value = self.calc_total_valuation(prior_winner_bundle)
                    prior_winner_bundle.remove(good)
                    curr_winner_value_minus_curr_item = self.calc_total_valuation(prior_winner_bundle)

                    if len(prior_winner_bundle) > self.min_items_in_bundle \
                         and (curr_winner_value_minus_curr_item == 0 
                              or curr_winner_value / curr_winner_value_minus_curr_item >= self.opponent_swap_factor):
                        self.play_aggressive[good] = True

                    # Verify if we want to swap to an aggressive strategy to secure ourself a better partition
                    curr_self_value = self.calc_total_valuation(tentative_allocation_goods)
                    tentative_allocation_goods.add(good)
                    curr_self_value_with_curr_item = self.calc_total_valuation(tentative_allocation_goods)

                    if len(prior_winner_bundle) > self.min_items_in_bundle \
                        and (curr_self_value == 0 
                             or curr_self_value_with_curr_item / curr_self_value >= self.self_swap_factor):
                        self.play_aggressive[good] = True

                    # Normal round strategy
                    if self.play_aggressive[good]:
                        max_valuation = valuations[good] * self.aggressive_bidding_factor_national
                        if min_bids[good] <= max_valuation:
                            bids[good] = random.random() * (max_valuation - min_bids[good]) + min_bids[good]
                    else:
                        max_valuation = valuations[good] * self.passive_bidding_factor
                        if min_bids[good] <= max_valuation:
                            bids[good] = min_bids[good]

        return bids

    def regional_bidder_strategy(self):
        min_bids = self.get_min_bids()
        valuations = self.get_valuations(self.get_goods())
        nearby_goods = self.get_goods_in_proximity()
        tentative_allocation_goods = self.get_tentative_allocation()
        prior_bids = self.get_previous_bid_map()

        bids = dict()

        # Iterate through each item
        for good in nearby_goods:
            if self.get_current_round() == 0:
                # Bid the initial bid to get information, or aggressively if we have a high value
                if valuations[good] >= self.distribution_threshold_regional:
                    self.play_aggressive[good] = True
                    bids[good] = self.aggressive_initial_jump_bid_factor * valuations[good]
                else:
                    self.play_aggressive[good] = False
                    bids[good] = min(self.initial_bid_regional, valuations[good])
            
            else:
                if self.get_current_round() == 1:
                    # Gather information about aggressive vs passive for each item right after first round
                    if good not in tentative_allocation_goods:
                        self.is_opponent_aggressive[good] = True
                    else:
                        self.is_opponent_aggressive[good] = False
                        self.play_aggressive[good] = True

                # If we already own the good, hold our bid
                if good in tentative_allocation_goods and good in prior_bids:
                    bids[good] = prior_bids[good]
                
                else:
                    # Verify if we want to swap to an aggressive strategy to block off an opponent
                    prior_winners = self.get_previous_winners_map()
                    curr_item_prior_winner = prior_winners[good]
                    prior_winner_bundle = set()

                    for winners_good in prior_winners:
                        if prior_winners[winners_good] == curr_item_prior_winner:
                            prior_winner_bundle.add(winners_good)
                    
                    curr_winner_value = self.calc_total_valuation(prior_winner_bundle)
                    prior_winner_bundle.remove(good)
                    curr_winner_value_minus_curr_item = self.calc_total_valuation(prior_winner_bundle)

                    if len(prior_winner_bundle) > self.min_items_in_bundle \
                         and (curr_winner_value_minus_curr_item == 0 
                              or curr_winner_value / curr_winner_value_minus_curr_item >= self.opponent_swap_factor):
                        self.play_aggressive[good] = True

                    # Verify if we want to swap to an aggressive strategy to secure ourself a better partition
                    curr_self_value = self.calc_total_valuation(tentative_allocation_goods)
                    tentative_allocation_goods.add(good)
                    curr_self_value_with_curr_item = self.calc_total_valuation(tentative_allocation_goods)

                    if len(prior_winner_bundle) > self.min_items_in_bundle \
                        and (curr_self_value == 0 
                             or curr_self_value_with_curr_item / curr_self_value >= self.self_swap_factor):
                        self.play_aggressive[good] = True

                    # Normal round strategy
                    if self.play_aggressive[good]:
                        max_valuation = valuations[good] * self.aggressive_bidding_factor_regional
                        if min_bids[good] <= max_valuation:
                            bids[good] = random.random() * (max_valuation - min_bids[good]) + min_bids[good]
                    else:
                        max_valuation = valuations[good] * self.passive_bidding_factor
                        if min_bids[good] <= max_valuation:
                            bids[good] = min_bids[good]

        return bids

    def get_bids(self):
        if self.is_national_bidder():
            return self.national_bidder_strategy()
        else:
            return self.regional_bidder_strategy()

    def update(self):
        pass
        # self.current_round = self.get_current_round()
        # self.current_prices = self.get_current_prices_map()
        # self.winner_history = self.get_winner_history_map()
        # self.recent_util = self.get_previous_util()
        # if not self.is_national:
        #     self.nearby_goods = self.get_goods_in_proximity()
        # self.adjust_bidding_strategy()

    # def proximity(self, goods, regional_good):
    #     # Correct handling of the proximity method
    #     return np.where(goods == regional_good)[0]

################### SUBMISSION #####################
my_agent_submission = MyAgent(NAME)
####################################################

# Additional functions and main definition as previously defined.

def process_saved_game(filepath): 
    """ 
    Here is some example code to load in a saved game in the format of a json.gz and to work with it
    """
    
    # NOTE: Data is a dictionary mapping 
    with gzip.open(filepath, 'rt', encoding='UTF-8') as f:
        game_data = json.load(f)
        for agent, agent_data in game_data.items(): 
            if agent_data['valuations'] is not None: 
                # agent is the name of the agent whose data is being processed 
                agent = agent 
                
                # bid_history is the bidding history of the agent as a list of maps from good to bid
                bid_history = agent_data['bid_history']
                
                # price_history is the price history of the agent as a list of maps from good to price
                price_history = agent_data['price_history']
                
                # util_history is the history of the agent's previous utilities 
                util_history = agent_data['util_history']
                
                # winner_history is the history of the previous tentative winners of all goods as a list of maps from good to winner
                winner_history = agent_data['winner_history']
                
                # elo is the agent's elo as a string
                elo = agent_data['elo']
                
                # is_national_bidder is a boolean indicating whether or not the agent is a national bidder in this game 
                is_national_bidder = agent_data['is_national_bidder']
                
                # valuations is the valuations the agent received for each good as a map from good to valuation
                valuations = agent_data['valuations']
                
                # regional_good is the regional good assigned to the agent 
                # This is None in the case that the bidder is a national bidder 
                regional_good = agent_data['regional_good']
            
            # TODO: If you are planning on learning from previously saved games enter your code below. 

if __name__ == "__main__":

    ### DO NOT TOUCH THIS #####
    agent = MyAgent(NAME)
    arena = LSVMArena(
        num_cycles_per_player = 3,
        timeout=1,
        local_save_path="saved_games",
        players=[
            agent,
            MyAgent("CP - MyAgent"),
            MyAgent("CP2 - MyAgent"),
            MyAgent("CP3 - MyAgent"),
            MinBidAgent("Min Bidder"), 
            JumpBidder("Jump Bidder"), 
            TruthfulBidder("Truthful Bidder"), 
        ]
    )

    start = time.time()
    arena.run()
    end = time.time()



from typing import Set, List, Dict, Iterable, Callable
import itertools
from abc import ABC, abstractmethod
from agt_server.agents.utils.adx.structures import Campaign, Bid
from math import isfinite, atan

class NDaysNCampaignsAgent(ABC):
    _uid_generator = itertools.count(1)

    uid: int
    def __init__(self):
        self.current_game = 0
        self.init()
        self.uid = next(type(self)._uid_generator)
        
    def init(self):
        self.current_day = 0
        self.quality_score = 1.0
        self.my_campaigns = set()
        self.statistics = {}
        self.profit = 0.0

    def __hash__(self):
        return hash(self.__class__.__name__ + str(self.uid))
    
    def get_quality_score(self) -> float:
        return self.quality_score

    def get_current_day(self) -> int:
        return self.current_day

    def current_game(self) -> int:
        return self.current_game

    def get_active_campaigns(self) -> Set[Campaign]:
        filtered = [c for c in self.my_campaigns if self.current_day >= c.start_day  and self.current_day <= c.end_day]
        return set(filtered)

    def get_cumulative_reach(self, campaign: Campaign) -> int:
        return campaign.cumulative_reach

    def get_cumulative_cost(self, campaign: Campaign) -> float:
         return campaign.cumulative_cost

    def get_cumulative_profit(self) -> float:
        return self.profit

    @staticmethod
    def effective_reach(x: int, R: int) -> float:
        return (2.0 / 4.08577) * (atan(4.08577 * ((x + 0.0) / R) - 3.08577) - atan(-3.08577))

    @staticmethod
    def is_valid_campaign_bid(campaign: Campaign, bid: float) -> bool:
        return isfinite(bid) and 0.1 * campaign.reach <= bid <= campaign.reach

    @staticmethod
    def clip_campaign_bid(campaign: Campaign, bid: float) -> float:
        return max(0.1 * campaign.reach, min(campaign.reach, bid))

    def effective_campaign_bid(self, bid: float) -> float:
        return bid / self.get_quality_score()

    @abstractmethod
    def get_ad_bids(self) -> Set[Bid]:
        pass

    @abstractmethod
    def get_campaign_bids(self, campaigns_for_auction: Set[Campaign]) -> Dict[Campaign, float]:
        pass

    @abstractmethod
    def on_new_game(self):
        pass

    def update_statistics(self, statistics):
        self.statistics.update(statistics)
    
    def __repr__(self):
        return self.name



import os

def delete_contents(path):
    for root, dirs, files in os.walk(path, topdown=False):  # Traverse from bottom-up to delete files first
        for file in files:
            if file == "__init__.py":
                continue
            file_path = os.path.join(root, file)
            os.remove(file_path)

        for dir_name in dirs:
            dir_path = os.path.join(root, dir_name)
            if not os.listdir(dir_path):  # If directory is empty, leave it intact
                continue


if __name__ == "__main__":
    logs_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "lsvm_logs_2025")
    if os.path.isdir(logs_dir):
        delete_contents(logs_dir)
        print(f"Cleared contents of '{logs_dir}' except for '__init__.py' files and folders.")
    else:
        print(f"Directory '{logs_dir}' does not exist.")



from abc import ABC, abstractmethod
from typing import Set, List, Dict
import itertools
from agt_server.agents.utils.adx.structures import Campaign
from agt_server.agents.base_agents.adx_agent import NDaysNCampaignsAgent



class State(ABC):
    _uid_generator = itertools.count(1)

    uid: int

    def __init__(self):
        super().__init__()
        self.uid = next(type(self)._uid_generator)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.uid == other.uid

    def __hash__(self):
        return hash(self.__class__.__name__ + str(self.uid))

    @abstractmethod
    def to_vector(self):
        return [self.uid]


class BidderState(State):
    _uid_generator = itertools.count(1)

    bidder: NDaysNCampaignsAgent

    def __init__(self, bidder):
        super().__init__()
        self.bidder = bidder

    def to_vector(self):
        return super().to_vector()


class CampaignBidderState(BidderState):
    _uid_generator = itertools.count(1)

    campaigns: Set[Campaign]
    budgets: Dict[Campaign, float]
    spend: Dict[Campaign, float]  # How much has been spent so far on each campaign (e.g., via bids won)
    impressions: Dict[Campaign, int]  # How many impressions have been acquired for each campaign
    quality_score: float
    timestep: int  # e.g., day of the AdX game

    def __init__(self, bidder, campaign: Campaign = None, budget: float = 0.0, spend: float = 0.0, impressions: int = 0, timestep: int = 0):
        super().__init__(bidder)
        self.campaigns = {campaign.uid: campaign} if campaign else {}
        self.budgets = {campaign.uid: budget} if campaign else {}
        self.spend = {campaign.uid: spend} if campaign else {}
        self.impressions = {campaign.uid: impressions} if campaign else {}
        self.expenditure = 0
        self.timestep = timestep
        self.profits = 0.0
        self.quality_score = 1.0

    def add_campaign(self, campaign):
        self.campaigns[campaign.uid] = campaign
        self.budgets[campaign.uid] = campaign.budget 
        self.spend[campaign.uid] = 0 
        self.impressions[campaign.uid] = 0 
    
    def add_all(self, campaigns):
        for c in campaigns:
            self.add_campaign(c)

    def __repr__(self):
        return "{}(campaigns: {}, budgets: {}, spend: {}, impressions: {}, timestep: {})".format(
            self.__class__.__name__,
            self.campaigns, self.budgets,
            self.spend, self.impressions,
            self.timestep
        )

    def to_vector(self):
        camp_vecs = []
        for campaign in self.campaigns:
            camp_vec = [campaign.reach, self.budgets[campaign], campaign.target.uid, self.spend[campaign], self.impressions[campaign]]
            camp_vecs.append(camp_vec)
        vec = [len(self.campaigns)] + [item for sublist in camp_vecs for item in sublist] + [self.timestep]
        return vec


import os
import importlib
import sys
import json
from datetime import datetime
import shutil
from agt_server.local_games.lsvm_arena import LSVMArena
import re

# Assuming the submissions directory is where the script is located
SUBMISSIONS_DIR = os.path.dirname(__file__)

import os

def add_init_files(directory):
    """
    Recursively creates __init__.py files in every subdirectory of the specified directory,
    ensuring that each directory is recognized as a Python package.
    Hidden directories (starting with '.') are ignored.
    """
    for root, dirs, _ in os.walk(directory):
        dirs[:] = [d for d in dirs if not d.startswith('.')]
        
        for dir in dirs:
            init_file_path = os.path.join(root, dir, '__init__.py')
            open(init_file_path, 'a').close()

def rename_directories(submissions_dir):
    for student_dir_name in os.listdir(submissions_dir):
        student_dir_path = os.path.join(submissions_dir, student_dir_name)
        if os.path.isdir(student_dir_path):
            new_name = student_dir_name.split('@')[0]  # Get the name before '@'
            new_dir_path = os.path.join(submissions_dir, new_name)
            os.rename(student_dir_path, new_dir_path)

def update_python_imports_to_relative(student_dir_path):
    """
    Updates Python files in the specified directory to use relative imports for
    modules that are part of the same submission, but only if the import matches
    a filename in the directory.
    """
    python_files = {file[:-3] for file in os.listdir(student_dir_path) if file.endswith('.py')}
    
    for root, dirs, files in os.walk(student_dir_path):
        for filename in files:
            if filename.endswith('.py'):
                file_path = os.path.join(root, filename)
                with open(file_path, 'r') as file:
                    lines = file.readlines()

                with open(file_path, 'w') as file:
                    for line in lines:
                        modified_line = line  
                        simple_import_match = re.match(r'^import (\w+)', line)
                        from_import_match = re.match(r'^from (\w+) import (\w+)', line)
                        if simple_import_match and simple_import_match.group(1) in python_files:
                            modified_line = f"from . import {simple_import_match.group(1)}\n"
                        elif from_import_match and from_import_match.group(1) in python_files:
                            modified_line = f"from .{from_import_match.group(1)} import {from_import_match.group(2)}\n"
                        file.write(modified_line)
    
def clear_directory_contents(student_dir_path):
    """
    Deletes specific files and directories within the specified directory
    and removes all print statements from Python files.
    """
    files_to_delete = ['.gitignore', 'local_elo_ranking.json']
    dirs_to_delete = ['saved_games']
    
    for filename in files_to_delete:
        filepath = os.path.join(student_dir_path, filename)
        if os.path.isfile(filepath):
            os.remove(filepath)
    
    for dirname in dirs_to_delete:
        dirpath = os.path.join(student_dir_path, dirname)
        if os.path.isdir(dirpath):
            shutil.rmtree(dirpath)
    
    # Remove print statements from Python files
    for root, _, files in os.walk(student_dir_path):
        for file in files:
            if file.endswith('.py'):
                file_path = os.path.join(root, file)
                with open(file_path, 'r') as f:
                    lines = f.readlines()
                
                with open(file_path, 'w') as f:
                    for line in lines:
                        if not re.match(r'\s*print\(.*\)', line):
                            f.write(line)
            
def clear_pycache(directory):
    """
    Recursively deletes __pycache__ directories within the specified directory.
    """
    for root, dirs, files in os.walk(directory):
        if '__pycache__' in dirs:
            shutil.rmtree(os.path.join(root, '__pycache__'))
            
def get_submission_modules(submissions_dir):
    submission_modules = []
    names_counter = {}  
    seen_imports = set()
    
    add_init_files(submissions_dir)  # Ensure all directories are packages
    rename_directories(submissions_dir)
    
    for student_dir_name in set(os.listdir(submissions_dir)):
        student_dir_path = os.path.join(submissions_dir, student_dir_name)
        if os.path.isdir(student_dir_path):
            clear_pycache(student_dir_path)
            for dirpath, _, filenames in os.walk(student_dir_path):
                clear_directory_contents(student_dir_path)
                if 'agent_submission.py' in filenames:
                    update_python_imports_to_relative(dirpath)  # Apply relative imports
                    relative_path = os.path.relpath(dirpath, SUBMISSIONS_DIR).replace(os.path.sep, '.')
                    import_path = f"{relative_path}.agent_submission"
                    break
            try:
                if import_path not in seen_imports: 
                    print(f"Import Path: {import_path}")
                    module = importlib.import_module(import_path)
                    agent_submission = getattr(module, 'agent_submission', None)
                    if agent_submission:
                        mod_time = os.path.getmtime(student_dir_path)
                        agent_submission.timestamp = mod_time
                        original_name = agent_submission.name
                        if original_name in names_counter:
                            names_counter[original_name] += 1
                            agent_submission.name = f"{original_name}({names_counter[original_name]})"
                        else:
                            names_counter[original_name] = 0
                        print(f"IMPORTING {agent_submission.name}")
                        submission_modules.append(agent_submission)
                        seen_imports.add(import_path)
                
            except Exception as e:
                print(f"Failed to import {student_dir_name}: {e}", file=sys.stderr)
            
    return submission_modules

if __name__ == "__main__":
    # Loading configuration from a JSON file
    config_path = os.path.join(SUBMISSIONS_DIR, "config.json")  # Adjust path as necessary
    with open(config_path) as cfile:
        server_config = json.load(cfile)
    
    print(SUBMISSIONS_DIR)
    agent_submissions = get_submission_modules(SUBMISSIONS_DIR)
    print(f"{len(agent_submissions)} successfully imported")
    print([agent.name for agent in agent_submissions if agent is not None])
    
    
    start = datetime.now()
    
    arena = LSVMArena(
            num_rounds = server_config['num_rounds'],
            num_cycles_per_player = 1, 
            timeout = server_config['response_time'],
            players = [agent for agent in agent_submissions if agent is not None],
            handin = True, 
            logging_path = server_config['logging_path'], 
            save_path = f"{server_config['save_path']}",
            local_save_path=f"{server_config['local_save_path']}"
        )
    # arena.set_ignores(["TA - Truthful", "TA - SECRET", "TA - RANDOM", "TA - MinBidder", "TA - Confidence", "TA - Jumper", "PW - Hot'n'Cold"])
    arena.run()
    end = datetime.now()
    print(f"Time Elapsed (s): {(end - start).total_seconds()}")



from pathlib import Path

def find_local_root(marker_file='.local_root') -> Path:
    """
    Find the project root directory marked by a specific file.

    Args:
        marker_file (str): The name of the marker file to look for.

    Returns:
        Path: The path to the project root directory.

    Raises:
        FileNotFoundError: If the project root cannot be found.
    """
    current_path = Path(__file__).resolve()

    for parent in current_path.parents:
        if (parent / marker_file).exists():
            return parent
    raise FileNotFoundError(f'Project root not found.')

def path_from_local_root(path = ''): 
    project_root = find_local_root()
    return project_root / path


import random
from typing import List, Dict, Optional
from agt_server.agents.utils.adx.structures import Campaign, Bid, BidBundle, MarketSegment
from agt_server.agents.utils.adx.pmfs import PMF
from agt_server.agents.utils.adx.states import CampaignBidderState
from collections import defaultdict
from agt_server.agents.base_agents.adx_agent import NDaysNCampaignsAgent
from math import isfinite, atan
import pkg_resources
from datetime import datetime
import os

CONFIG = {
        'num_agents': 10,
        'num_days': 10,
        'quality_score_alpha': 0.5,
        'campaigns_per_day': 5,
        'campaign_reach_dist': [0.3,0.5,0.7],
        'campaign_length_dist': [1, 2, 3],
        'market_segment_dist': [
            MarketSegment(("Male", "Young")),
            MarketSegment(("Male", "Old")),
            MarketSegment(("Male", "LowIncome")),
            MarketSegment(("Male", "HighIncome")),
            MarketSegment(("Female", "Young")),
            MarketSegment(("Female", "Old")),
            MarketSegment(("Female", "LowIncome")),
            MarketSegment(("Female", "HighIncome")),
            MarketSegment(("Young", "LowIncome")),
            MarketSegment(("Young", "HighIncome")),
            MarketSegment(("Old", "LowIncome")),
            MarketSegment(("Old", "HighIncome")),
            MarketSegment(("Male", "Young", "LowIncome")),
            MarketSegment(("Male", "Young", "HighIncome")),
            MarketSegment(("Male", "Old", "LowIncome")),
            MarketSegment(("Male", "Old", "HighIncome")),
            MarketSegment(("Female", "Young", "LowIncome")),
            MarketSegment(("Female", "Young", "HighIncome")),
            MarketSegment(("Female", "Old", "LowIncome")),
            MarketSegment(("Female", "Old", "HighIncome"))
        ],
        'market_segment_pop': {
            MarketSegment(("Male", "Young")): 2353,
            MarketSegment(("Male", "Old")): 2603,
            MarketSegment(("Male", "LowIncome")): 3631,
            MarketSegment(("Male", "HighIncome")): 1325,
            MarketSegment(("Female", "Young")): 2236,
            MarketSegment(("Female", "Old")): 2808,
            MarketSegment(("Female", "LowIncome")): 4381,
            MarketSegment(("Female", "HighIncome")): 663,
            MarketSegment(("Young", "LowIncome")): 3816,
            MarketSegment(("Young", "HighIncome")): 773,
            MarketSegment(("Old", "LowIncome")): 4196,
            MarketSegment(("Old", "HighIncome")): 1215,
            MarketSegment(("Male", "Young", "LowIncome")): 1836,
            MarketSegment(("Male", "Young", "HighIncome")): 517,
            MarketSegment(("Male", "Old", "LowIncome")): 1795,
            MarketSegment(("Male", "Old", "HighIncome")): 808,
            MarketSegment(("Female", "Young", "LowIncome")): 1980,
            MarketSegment(("Female", "Young", "HighIncome")): 256,
            MarketSegment(("Female", "Old", "LowIncome")): 2401,
            MarketSegment(("Female", "Old", "HighIncome")): 407
        },
        'user_segment_pmf': {
            MarketSegment(("Male", "Young", "LowIncome")): 0.1836,
            MarketSegment(("Male", "Young", "HighIncome")): 0.0517,
            MarketSegment(("Male", "Old", "LowIncome")): 0.1795,
            MarketSegment(("Male", "Old", "HighIncome")): 0.0808,
            MarketSegment(("Female", "Young", "LowIncome")): 0.1980,
            MarketSegment(("Female", "Young", "HighIncome")): 0.0256,
            MarketSegment(("Female", "Old", "LowIncome")): 0.2401,
            MarketSegment(("Female", "Old", "HighIncome")): 0.0407
        }
    }

def calculate_effective_reach(x: int, R: int) -> float:
    return (2.0 / 4.08577) * (atan(4.08577 * ((x + 0.0) / R) - 3.08577) - atan(-3.08577))


class AdXGameSimulator:
    def __init__(self, config: Optional[Dict] = None, save_path = None, local_save_path = None):
        if config is None:
            config = CONFIG
        self.save_path = save_path
        self.local_save_path = local_save_path
        self.num_agents = config['num_agents']
        self.num_days = config['num_days']
        self.α = config['quality_score_alpha']
        self.campaigns_per_day = config['campaigns_per_day']
        self.agents: List = []
        self.campaign_reach_dist = config['campaign_reach_dist']
        self.campaign_length_dist = config['campaign_length_dist']
        self.market_segment_dist = config['market_segment_dist']
        self.market_segment_pop = config['market_segment_pop']
        self.user_segment_dist = PMF(config['user_segment_pmf'])
        self.sub_segments = defaultdict(list)
        for user_seg in config['user_segment_pmf']:
            for market_seg in config['market_segment_dist']:
                if market_seg.issubset(user_seg):
                    self.sub_segments[user_seg].append(market_seg)
        
        # Set up log file if local_save_path is provided
        if self.local_save_path:
            timestamp = datetime.now().strftime('%Y-%m-%d_%H%M%S')
            self.log_file = os.path.join(self.local_save_path, f"{timestamp}.txt")
        else:
            self.log_file = None

    def log(self, message: str) -> None:
        """Log messages to console and file if specified."""
        if self.log_file:
            with open(self.log_file, 'a') as f:
                f.write(message + '\n')
        else: 
            print(message)

    def init_agents(self, agent_types: List[type]) -> Dict[NDaysNCampaignsAgent, CampaignBidderState]:
        states = dict()
        self.agents = []
        for i, agent_type in enumerate(agent_types):
            try:
                agent_type.init()
                self.agents.append(agent_type)
                self.agents[i].agent_num = i
                states[agent_type] = CampaignBidderState(i)
            except Exception as e:
                self.log(f"Error initializing agent {agent_type}: {e}")
                continue
        return states 
    
    def generate_campaign(self, start_day: int, end_day: Optional[int] = None) -> Campaign:
        delta = random.choice(self.campaign_reach_dist)
        length =  random.choice(self.campaign_length_dist)
        mkt_segment = random.choice(self.market_segment_dist)
        reach = int(self.market_segment_pop[mkt_segment] * delta)
        if end_day is None:
            end_day=start_day + length - 1
        return Campaign(reach=reach, 
                        target=mkt_segment, 
                        start_day=start_day, 
                        end_day=end_day)
    
    def is_valid_campaign_bid(self, bid: float, reach: int) -> bool:
        return isfinite(bid) and 0.1 * reach <= bid <= reach
        
    def is_valid_bid(self, bid: Bid) -> bool:
        bid = bid.bid_per_item
        return isfinite(bid) and bid > 0

    def run_ad_auctions(self, bid_bundles : List[BidBundle], users: List[MarketSegment], day: int) -> None:
        bidder_states = self.states
        seg_to_bid = defaultdict(set)
        daily_limits = dict()
        bid_to_bundle = dict()
        bid_to_spend = dict()

        for bid_bundle in bid_bundles:
            try:
                daily_limits[bid_bundle.campaign_id] = bid_bundle.limit
                if bid_bundle.campaign_id not in self.campaigns:
                    continue
                campaign = self.campaigns[bid_bundle.campaign_id]
                if not (campaign.start_day <= day and campaign.end_day >= day):
                    continue

                for bid in bid_bundle.bid_entries:
                    if self.is_valid_bid(bid):
                        bid_to_bundle[bid] = bid_bundle 
                        seg_to_bid[bid.item].add(bid)
                        bid_to_spend[bid] = 0
            except Exception as e:
                self.log(f"Error processing bid bundle {bid_bundle}: {e}")
                continue

        for user_segment in users:
            try:
                bids = []
                sub_segments = self.sub_segments[user_segment]
                for seg in sub_segments:
                    bids.extend(seg_to_bid[seg])

                bids.sort(key=lambda b: b.bid_per_item, reverse=True)

                for i, bid in enumerate(bids):
                    try:
                        campaign_id = bid_to_bundle[bid].campaign_id
                        price = bids[i + 1].bid_per_item if i + 1 < len(bids) else 0     
                        bidder_states[bid.bidder].spend[campaign_id] += price
                        over_bid_limit = bid_to_spend[bid] + price > bid.bid_limit
                        over_bundle_limit = bidder_states[bid.bidder].spend[campaign_id] + price > daily_limits[campaign_id]
                        
                        if over_bid_limit or over_bundle_limit:
                            seg_to_bid[bid.item].remove(bid)
                            continue 
                        else: 
                            bid_to_spend[bid] += price
                            bidder_state = bidder_states[bid.bidder]
                            campaign_id = bid_to_bundle[bid].campaign_id
                            campaign = bidder_state.campaigns[campaign_id]
                            if campaign is not None:
                                bidder_state.spend[campaign_id] += price
                                campaign.cumulative_cost += price
                                if campaign.target_segment.issubset(user_segment):
                                    bidder_state.impressions[campaign_id] += 1
                                    campaign.cumulative_reach += 1 
                            break 
                    except Exception as e:
                        self.log(f"Error processing bid {bid}: {e}")
                        continue
            except Exception as e:
                self.log(f"Error processing user segment {user_segment}: {e}")
                continue


    def run_campaign_auctions(self, agent_bids: Dict[NDaysNCampaignsAgent, Dict[Campaign, float]], new_campaigns: List[Campaign]) -> None:
        new_campaigns = set(new_campaigns)
        for campaign in new_campaigns:
            bids = []
            for agent in self.agents:
                try:
                    if campaign in agent_bids[agent]:
                        reach = campaign.reach
                        agent_bid = agent_bids[agent][campaign]
                        if self.states[agent].quality_score > 0 and self.is_valid_campaign_bid(agent_bid, reach):
                            effective_bid = agent_bid / self.states[agent].quality_score
                            bids.append((agent, effective_bid))
                except Exception as e:
                    self.log(f"Error processing campaign auction for agent {agent}: {e}")
                    continue
                   
            if bids:
                try:
                    winner, effective_bid = min(bids, key=lambda x: x[1])
                    if len(bids) == 1:
                        q_low = 0.0
                        all_quality_scores = [self.states[agent].quality_score for agent in self.agents]
                        sorted_quality_scores = sorted(all_quality_scores)
                        for i in range(3):
                            q_low += sorted_quality_scores[min(i, len(sorted_quality_scores) - 1)]
                        q_low /= 3
                        budget = campaign.reach / q_low * self.states[winner].quality_score
                    else:
                        second_lowest_bid = sorted(bids, key=lambda x: x[1])[1][1]
                        budget = second_lowest_bid * self.states[winner].quality_score
                    campaign.budget = budget
                    winner.my_campaigns.add(campaign)
                    winner_state = self.states[winner]
                    winner_state.add_campaign(campaign)
                    self.campaigns[campaign.uid] = campaign
                except Exception as e:
                    self.log(f"Error processing campaign auction results for campaign {campaign.uid}: {e}")


    def generate_auction_items(self, num_items: int) -> List[MarketSegment]:
        return [item for item in self.user_segment_dist.draw_n(num_items, replace=True)]

    def print_day_summary(self, day: int) -> None:
        self.log(f"\n\t################# DAY {day} SUMMARY ##################")
        self.log("\n\t#### Agent \t\t# Profit \t###")
        self.log("\n\t###########################################")
        for agent in self.agents:
            agent_state = self.states[agent]
            self.log(f"\n\t### {agent.name:12} \t# {agent_state.profits:8.2f} \t###")
        self.log("\n\t###########################################")

    def print_game_results(self) -> None:
        self.log("\n\t################# SIMULATION RESULTS ##################")
        self.log("\n\t#### Agent \t\t# Profit \t###")
        self.log("\n\t###########################################")
        for agent in sorted(self.agents, key=lambda a: self.states[a].profits, reverse=True):
            agent_state = self.states[agent]
            self.log(f"\n\t### {agent.name:12} \t# {agent_state.profits:8.2f} \t###")
        self.log("\n\t###########################################")
        winner = max(self.agents, key=lambda a: self.states[a].profits)
        self.log(f"\n\t@@@ WINNER: {winner.name} @@@")
    
    def print_final_results(self, total_profits: Dict[NDaysNCampaignsAgent, float], num_simulations: int) -> None:
        self.log("\n\t################# FINAL SIMULATION RESULTS ##################")
        self.log("\n\t#### Agent \t\t# Profit \t###")
        self.log("\n\t###########################################")
        
        # Collect results for the final string
        result_list = []
        
        for agent in sorted(self.agents, key=lambda a: total_profits[a.name], reverse=True):
            avg_profit = total_profits[agent.name] / num_simulations
            self.log(f"\n\t### {agent.name:12} \t# {avg_profit:8.2f} \t###")
            
            # Append to the result list
            result_list.append(agent.name)
            result_list.append(f"{avg_profit:.2f}")
        
        self.log("\n\t###########################################")
        winner = max(self.agents, key=lambda a: total_profits[a.name])
        self.log(f"\n\t@@@ WINNER: {winner.name} @@@")
        
        if self.save_path:
            # Format today's date
            today_date_formatted = datetime.now().strftime('%Y-%m-%d_%H%M')
            
            # Prepare the final string
            final_str = f"{today_date_formatted}\t" + "\t".join(result_list)
            
            # Log the final string for confirmation
            self.log(final_str)
            
            # Save to file
            with open(pkg_resources.resource_filename('agt_server', self.save_path), 'a') as file:
                file.write(final_str + "\n")
                

    def run_simulation(self, agents: list[NDaysNCampaignsAgent], num_simulations: int) -> None:
        total_profits = {agent.name : 0.0 for agent in agents}
        for i in range(num_simulations):
            self.states = self.init_agents(agents)
            self.campaigns = dict()
            # Initialize campaigns 
            for agent in self.agents:
                try:  
                    agent.current_game = i + 1 
                    agent.my_campaigns = set()
                    agent.on_new_game()
                    random_campaign = self.generate_campaign(start_day=1)
                    agent_state = self.states[agent]
                    random_campaign.budget = random_campaign.reach
                    agent_state.add_campaign(random_campaign)
                    agent.my_campaigns.add(random_campaign)
                    self.campaigns[random_campaign.uid] = random_campaign
                except Exception as e:
                    self.log(f"Error initializing agent {agent}: {e}")
                    continue

            for day in range(1, self.num_days + 1):
                # Update 
                for agent in self.agents:
                    try:
                        agent.current_day = day
                    except Exception as e:
                        self.log(f"Error updating agent {agent} for day {day}: {e}")
                        continue

                # Generate new campaigns and filter
                if day + 1 < self.num_days + 1:
                    new_campaigns = [self.generate_campaign(start_day=day + 1) for _ in range(self.campaigns_per_day)]
                    new_campaigns = [c for c in new_campaigns if c.end_day <= self.num_days]
                    # Solicit campaign bids and run campaign auctions
                    agent_bids = dict()
                    for agent in self.agents:
                        try:
                            agent_bids[agent] = agent.get_campaign_bids(new_campaigns)
                        except Exception as e:
                            self.log(f"Error getting campaign bids from agent {agent}: {e}")
                            agent_bids[agent] = {}
                            continue

                # Solicit ad bids from agents and run ad auctions
                ad_bids = []
                for agent in self.agents:
                    try:
                        ad_bids.extend(agent.get_ad_bids())
                    except Exception as e:
                        self.log(f"Error getting ad bids from agent {agent}: {e}")
                        continue
                users = self.generate_auction_items(10000)
                self.run_ad_auctions(ad_bids, users, day)

                # Update campaign states, quality scores, and profits
                for agent in self.agents:
                    agent_state = self.states[agent]
                    todays_profit = 0.0
                    new_qs_count = 0
                    new_qs_val = 0.0

                    for campaign in agent_state.campaigns.values():
                        if campaign.start_day <= day <= campaign.end_day:
                            if day == campaign.end_day:
                                impressions = agent_state.impressions[campaign.uid]
                                total_cost = agent_state.spend[campaign.uid]
                                effective_reach = calculate_effective_reach(impressions, campaign.reach)
                                todays_profit += (effective_reach) * agent_state.budgets[campaign.uid] - total_cost

                                new_qs_count += 1
                                new_qs_val += effective_reach

                    if new_qs_count > 0:
                        try:
                            new_qs_val /= new_qs_count
                            self.states[agent].quality_score = (1 - self.α) * self.states[agent].quality_score + self.α * new_qs_val
                            agent.quality_score = self.states[agent].quality_score
                        except Exception as e:
                            self.log(f"Error updating quality score for agent {agent}: {e}")
                            self.states[agent].quality_score = 0.0
                            agent.quality_score = 0.0

                    agent_state.profits += todays_profit
                    agent.profit += todays_profit
                
                # Run campaign auctions
                self.run_campaign_auctions(agent_bids, new_campaigns)
                # Run campaign endowments
                for agent in self.agents:
                    try:
                        if random.random() < min(1, agent.quality_score):
                            random_campaign = self.generate_campaign(start_day=day)
                            agent_state = self.states[agent]
                            random_campaign.budget = random_campaign.reach
                            agent_state.add_campaign(random_campaign)
                            agent.my_campaigns.add(random_campaign)
                            self.campaigns[random_campaign.uid] = random_campaign
                    except Exception as e:
                        self.log(f"Error running campaign endowments for agent {agent}: {e}")
                        continue
            
            for agent in self.agents:
                try:
                    total_profits[agent.name] += self.states[agent].profits 
                except Exception as e:
                    self.log(f"Error updating total profits for agent {agent}: {e}")
                    continue
                
            self.print_game_results()
        self.print_final_results(total_profits, num_simulations)
        return total_profits



import random
from typing import Set, Dict
from agt_server.agents.utils.adx.structures import Bid, Campaign, BidBundle, MarketSegment
from agt_server.agents.base_agents.adx_agent import NDaysNCampaignsAgent

class Tier1NDaysNCampaignsAgent(NDaysNCampaignsAgent):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def get_ad_bids(self) -> Set[BidBundle]:
        bundles = set()
        for campaign in self.get_active_campaigns():
            bids = set()
            bid_per_item = min(1, max(0.1, (campaign.budget - self.get_cumulative_cost(campaign)) /
                               (campaign.reach - self.get_cumulative_reach(campaign) + 0.0001)))
            total_limit = max(1.0, campaign.budget - self.get_cumulative_cost(campaign))
            auction_item = campaign.target_segment
            bid = Bid(self, auction_item, bid_per_item, total_limit)
            bids.add(bid)
            bundle = BidBundle(campaign_id=campaign.uid, limit=total_limit, bid_entries=bids)
            bundles.add(bundle)
        return bundles

    def get_campaign_bids(self, campaigns_for_auction: Set[Campaign]) -> Dict[Campaign, float]:
        bids = {}
        for campaign in campaigns_for_auction:
            bid_value = campaign.reach * (random.random() * 0.9 + 0.1)
            bids[campaign] = bid_value
        return bids

    def on_new_game(self):
        pass

################### SUBMISSION #####################
my_agent_submission = Tier1NDaysNCampaignsAgent("TA - Tier 1 Agent")
####################################################